#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2015 Michel Barbeau.
# Version: August 23, 2015 

import numpy
from gnuradio import gr
import random
from operator import xor
import pmt # polymorphic data types

class decoder(gr.sync_block):
    "Quantum decoder"
    # rectilinear basis & diagonal basis angles
    basis = [[0, 90], [45, 135]]
    # received key bit buffer
    key_bit = []
    # selected random basis
    sel_basis = -1  # first item is not used, set to impossible to match value
    # decoded bit
    dec_bit = 0

    def __init__(self):
        gr.sync_block.__init__(
            self,
            name = "decoder",
            # input signature
	    # in port of photon angles: 0s, 90s, 45s and 135s 
            in_sig = [numpy.uint32,
	    # in port of selected bases: 0s and 1s
		      numpy.uint32],
	    # output signature
	    # out port for plain text 
            out_sig = None)
	# asynchronous encrypted data port, 0s and 1s
 	self.message_port_register_in(pmt.intern('ciphertext'))
        self.set_msg_handler(pmt.intern('ciphertext'), self.handle_msg)
	# asynchronous to-encoder feedback port, 0s (neg.) and 1s (pos.) 
 	self.message_port_register_out(pmt.intern('feedback'))

    # from-encoder encrypted data port handler
    def handle_msg(self, msg):
        # consume encrypted data from encoder
	ciphertext = pmt.to_python(msg)
	for i in range(len(ciphertext)):
		# use one key bit
		k = self.key_bit.pop(0)
		print "decoder:data: "+str(ciphertext[i]^k)


    def work(self, input_items, output_items):
	# feedback to encoder
	feedback = []
	# read photon angle, previous basis pairs
	for i in range(len(input_items[0])):
		# compares encoder and decoder basis used for previous photon
		if input_items[0][i]==self.sel_basis:
			# append positive feedback
			feedback.append(1)
			# append corresponding biy to key
			self.key_bit.append(self.dec_bit)
		else:
			# append negative feedback
			feedback.append(0)

		# randomly select a basis
        	self.sel_basis=random.randint(0, 1)
		# decode a photon
		if input_items[1][i]==self.basis[self.sel_basis][0]:
			# store a zero
			self.dec_bit = 0
		elif input_items[1][i]==self.basis[self.sel_basis][1]:
			# store a one
			self.dec_bit = 1
		else:
			# store a random value
			self.dec_bit = random.randint(0, 1)
        # post feedback to encoder
        self.message_port_pub(pmt.intern('feedback'), 
		pmt.to_pmt(feedback)) 
	return len(input_items[0])


