#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2015 Michel Barbeau.
# Version: August 23, 2015 

import numpy
from gnuradio import gr
import random
import pmt # polymorphic data types

class encoder(gr.interp_block):
    "Quantum encoder"
    # rectilinear basis & diagonal basis angles
    basis = [[0, 90], [45, 135]]
    # plain text buffer
    plain_text = []
    # from-decoder feedback 
    feedback = []
    # potential key bit buffer
    key_bit = [0] # first item is random and not used (feedback always neg)
    # selected random bases
    sel_basis = 0 # first item is random and not used

    # constructor
    def __init__(self):
        gr.interp_block.__init__(
            self,
            name="encoder",
	    # input signature
	    # in port for plain text 
            in_sig = [numpy.uint32],	
            # output signature
	    # out port of photon angles: 0s, 90s, 45s and 135s 
            out_sig = [numpy.uint32,
	    # out port of selected bases: 0s and 1s
                       numpy.uint32],
	    interp = 2) # interpolation ratio
	# asynchronous from-decoder feedback port, 0s (neg.) and 1s (pos.) 
 	self.message_port_register_in(pmt.intern('feedback'))
        self.set_msg_handler(pmt.intern('feedback'), self.handle_msg)
	# asynchronous encrypted data port, 0s and 1s
 	self.message_port_register_out(pmt.intern('ciphertext'))

    # from-decoder feedback port handler
    def handle_msg(self, msg):
	# save the from-decoder feedback
	self.feedback.extend(pmt.to_python(msg))
	# encrypted data
	ciphertext = []
        # consume feedback from encoder
	for _ in range(min(len(self.feedback),len(self.plain_text))):
		# consume a feedback item
		f = self.feedback.pop(0)
		print "encoder:feedback: "+str(f)
		# consume a key bit
		k = self.key_bit.pop(0)
		if f: # positive feedback?
			# get a data bit
			d = self.plain_text.pop(0)
			print "encoder:data: "+str(d)
			# post encrypted data bit
			ciphertext.append(k^d)
        # post encrypted data
        self.message_port_pub(pmt.intern('ciphertext'),
           pmt.to_pmt(ciphertext))

    # main
    def work(self, input_items, output_items):
	# buffer input data
	self.plain_text.extend(input_items[0])
	# post photon angle, previous basis pairs
	for i in range(len(output_items[0])):
		# set basis used for previous photon
		output_items[0][i] = self.sel_basis
		# randomly select a bit
        	r=random.randint(0, 1)
		# randomly select a basis
        	self.sel_basis = random.randint(0, 1)
		# set angle corresponding to "sel_basis" and "r"
		output_items[1][i] = self.basis[self.sel_basis][r]
		# memorize random bit
		self.key_bit.append(r)
	print "encoder:bases: "+str(output_items[0])+\
		" angles: "+str(output_items[1])
	# return number of posted pairs
	return len(output_items[0])
	

